/* ///////////////////////////////////////////////////////////////////// */
/*!
  \file
  \brief Kelvin-Helmholtz instability setup.

  We consider a simple 2D, shear layer configuration, with constant
  density and pressure.
  The velocity profile is simply given by \f[ v_y = M/2*tanh(y/w)  \f], where
  M is the relative Mach number, w is the shear width.
  
  For a magnetized plasma, two additional parameters, namely the Alfven
  velocity (vA) and the angle theta between the magnetic field and
  the x-direction are used to control the field strength and orientation.

  Velocities are normalized to the sound speed, density to the fluid
  initial density and length to the total domain size.

  A passive tracer is used to distinguish the two fluids (+1 and -1 for the
  fluid above and below, respectively).

  Input parameters are:
  
  - g_inputParam[AMPL] = perturbation amplitude
  - g_inputParam[MACH] = relative velocity (and Mach number)
                         between the two layers;
  - g_inputParam[WIDTH] = width of the shear;

  For MHD, 
 
  - g_inputParam[VALF]  = Alfven velocity in units of the sound speed (=1)
  - g_inputParam[THETA] = angle (in degrees) between the magnetic field 
                          and the x direction.

  \authors A. Mignone (mignone@to.infn.it)
  \date    Aug 5, 2019
*/
/* ///////////////////////////////////////////////////////////////////// */
#include "pluto.h"

/* ********************************************************************* */
void Init (double *v, double x, double y, double z)
/*
 *
 *********************************************************************** */
{
  static int first_call = 1;
  double B0, rnd, scrh;
  double Ly    = g_domEnd[JDIR] - g_domBeg[IDIR];
  double M     = g_inputParam[MACH];
  double eps   = g_inputParam[AMPL]*M;  /* Perturbation amplitude */
  double a     = g_inputParam[WIDTH];
  double vA    = g_inputParam[VALF];
  double theta = g_inputParam[THETA]/180.0*CONST_PI;
 
  if (first_call == 1){          /* Seed random number sequence */
    RandomSeed(prank,0);   
    first_call = 0;
  }
  rnd  = RandomNumber(-1,1);         /* Generate random number in (-1,1) */
  scrh = y/(0.25*Ly);

  v[RHO] = 1.0;                     /* Set constant density  */ 
  v[VX1] = 0.5*M*tanh(y/a);         /* Main flow             */ 
  v[VX2] = eps*rnd*exp(-scrh*scrh);  /* Vertical perturbation */
  v[VX3] = 0.0;

  #if EOS == IDEAL
  v[PRS] = 1.0/g_gamma;             /* Set constant pressure (cs = 1) */
  #elif EOS == ISOTHERMAL
  g_isoSoundSpeed = 1.0;
  #endif
  v[TRC]  = (y < 0.0 ? 1.0:-1.0);   /* Passive tracer (color)         */ 

  #if PHYSICS == MHD
  B0     = vA*sqrt(v[RHO]);
  v[BX1] = B0*cos(theta);    /* Magnetic field lies in the x-z plane.        */
  v[BX2] = 0.0;              /* Theta is the angle between B and x-direction */
  v[BX3] = B0*sin(theta); 

  v[AX1] = 0.0;
  v[AX2] = 0.0;
  v[AX3] = y*v[BX1];
  #endif
}

/* ********************************************************************* */
void InitDomain (Data *d, Grid *grid)
/*! 
 * Assign initial condition by looping over the computational domain.
 * Called after the usual Init() function to assign initial conditions
 * on primitive variables.
 * Value assigned here will overwrite those prescribed during Init().
 *
 *
 *********************************************************************** */
{
}

/* ********************************************************************* */
void Analysis (const Data *d, Grid *grid)
/* 
 *
 *
 *********************************************************************** */
{
  int    i,j,k;  
  double max_vy  = -10.0;
  double min_vy  =  10.0;
  double max_By2 = 0.0;
  double pm = 0.0;
  double scrh;
  double ***vy = d->Vc[VX2];
  #if PHYSICS == MHD
  double ***Bx = d->Vc[BX1];
  double ***By = d->Vc[BX2];
  #endif
  double *dx = grid->dx[IDIR];
  double *dy = grid->dx[JDIR];
  double vol =  (g_domEnd[IDIR] - g_domBeg[IDIR])
               *(g_domEnd[JDIR] - g_domBeg[JDIR]);

  DOM_LOOP(k,j,i){
    max_vy = MAX(max_vy, vy[k][j][i]);
    min_vy = MIN(min_vy, vy[k][j][i]);
    
    #if PHYSICS == MHD
    scrh    = By[k][j][i]*By[k][j][i];
    max_By2 = MAX(max_By2, scrh);

    pm += 0.5*(Bx[k][j][i]*Bx[k][j][i] + By[k][j][i]*By[k][j][i])*dx[i]*dy[j];
    #endif
  }

/* -- parallel reduce -- */

  #ifdef PARALLEL
  MPI_Allreduce (&max_vy, &scrh, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
  max_vy = scrh;

  MPI_Allreduce (&min_vy, &scrh, 1, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
  min_vy = scrh;

  MPI_Allreduce (&max_By2, &scrh, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
  max_By2 = scrh;

  MPI_Allreduce (&pm, &scrh, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  pm = scrh/vol;

  MPI_Barrier (MPI_COMM_WORLD);
  #endif

/* --  Proc #0 writes to disk -- */

  if (prank == 0){
    FILE *fp;
    char *legend[4] = {"t","max-min","max(By^2)","pm"};
    if (g_stepNumber == 0) {
      fp = fopen("kh.dat","w");
      PrintColumnLegend (legend,4,fp);
    } else fp = fopen("kh.dat","a");

    fprintf (fp, "%12.6e  %12.6e  %12.6e  %12.6e\n",g_time, 
                                             0.5*(max_vy - min_vy),
                                             max_By2, pm);
    fclose(fp);
  }

}

/* ********************************************************************* */
void UserDefBoundary (const Data *d, RBox *box, int side, Grid *grid) 
/*! 
 *
 *
 *********************************************************************** */
{ }

